"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from SlurmDagman.config.data.package import package_config
from SlurmDagman.config.defaults.common import DEFAULTS as common_config_defaults
from SlurmDagman.config.defaults.process import DEFAULTS as process_config_defaults
from SlurmDagman.config.manager import ConfigManager

process_config = ConfigManager()
process_config.set_params(common_config_defaults)
process_config.set_params(package_config.get_params(common_config_defaults))
process_config.set_params(process_config_defaults)



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

DEFAULTS = OrderedDict()
DEFAULTS['DAGMAN'] = OrderedDict()
DEFAULTS['DAGMAN']['sleep_time'] = '120'
DEFAULTS['DAGMAN']['max_jobs_queued'] = '500'
DEFAULTS['DAGMAN']['max_jobs_submit'] = '0'
DEFAULTS['DAGMAN']['submit_wait_time'] = '2'



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import math
import os

from SlurmDagman.config.data.package import package_config
from SlurmDagman.dag import Dag


class CondorToSlurmTranslator(object):

    def __init__(self, condor_dag_file=None, slurm_dag_file=None,
                       slurm_partition=None, slurm_qos=None, slurm_time_limit=None, slurm_scratch_dir=None, slurm_use_setsid=None,
                       singularity_image=None, singularity_bind_mount_dbus=None):
        super(CondorToSlurmTranslator, self).__init__()
        self.condor_default_cpus = '1'
        self.condor_default_memory = '1GB'
        self.__init_dags(condor_dag_file, slurm_dag_file)
        self.__init_params()
        self.set_params(slurm_partition, slurm_qos, slurm_time_limit, slurm_scratch_dir, slurm_use_setsid,
                        singularity_image, singularity_bind_mount_dbus)


    def __init_dags(self, condor_dag_file=None, slurm_dag_file=None):
        self.__init_condor_dag(condor_dag_file)
        self.__init_slurm_dag(slurm_dag_file)


    def __init_condor_dag(self, condor_dag_file=None):
        self.condor_dag = Dag()
        if condor_dag_file:
            self.set_condor_dag_file(condor_dag_file)


    def __init_slurm_dag(self, slurm_dag_file=None):
        self.slurm_dag = Dag()
        if slurm_dag_file:
            self.set_slurm_dag_file(slurm_dag_file)


    def reset_dags(self):
        self.reset_condor_dag()
        self.reset_slurm_dag()


    def reset_condor_dag(self):
        self.__init_condor_dag()


    def reset_slurm_dag(self):
        self.__init_slurm_dag()


    def set_condor_dag_file(self, condor_dag_file):
        self.condor_dag.set_dag_file(condor_dag_file)


    def set_slurm_dag_file(self, slurm_dag_file):
        self.slurm_dag.set_dag_file(slurm_dag_file)


    def __init_params(self):
        self.slurm_partition = package_config.get_param('SLURM', 'partition')
        self.slurm_qos = package_config.get_param('SLURM', 'qos')
        self.slurm_time_limit = package_config.get_param('SLURM', 'time_limit')
        self.slurm_scratch_dir = package_config.get_param('SLURM', 'scratch_dir')
        self.slurm_use_setsid = False
        self.singularity_image = ''
        self.singularity_bind_mount_dbus = False


    def reset_params(self):
        self.__init_params()


    def set_params(self, slurm_partition=None, slurm_qos=None, slurm_time_limit=None, slurm_scratch_dir=None, slurm_use_setsid=None,
                         singularity_image=None, singularity_bind_mount_dbus=None):
        if slurm_partition is not None:
            self.slurm_partition = slurm_partition
        if slurm_qos is not None:
            self.slurm_qos = slurm_qos
        if slurm_time_limit is not None:
            self.slurm_time_limit = slurm_time_limit
        if slurm_scratch_dir is not None:
            self.slurm_scratch_dir = slurm_scratch_dir
        if slurm_use_setsid is not None:
            self.slurm_use_setsid = slurm_use_setsid
        if singularity_image is not None:
            self.singularity_image = singularity_image
        if singularity_bind_mount_dbus is not None:
            self.singularity_bind_mount_dbus = singularity_bind_mount_dbus


    def translate(self):
        self.translate_dag_file()
        self.translate_job_submission_files()


    def translate_dag_file(self):
        self.__parse_condor_dag_file()
        self.__condor_to_slurm_dag()
        self.__write_slurm_dag_file()


    def __parse_condor_dag_file(self):
        self.condor_dag.parse()


    def __condor_to_slurm_dag(self):
        self.slurm_dag.set_dag(self.condor_dag, True)
        for node in self.slurm_dag:
            self.slurm_dag[node]['job_submission_file'] = self.condor_dag[node]['job_submission_file'] + '.slurm'
            self.slurm_dag[node]['vars'] = self.slurm_dag[node]['vars'].replace('_CONDOR_SCRATCH_DIR', self.slurm_scratch_dir)


    def __write_slurm_dag_file(self):
        self.slurm_dag.write(use_dag_nodes_appearance_order=True)


    def translate_job_submission_files(self):
        condor_job_submission_files = []
        for dag_node in self.condor_dag:
            condor_job_submission_file = self.condor_dag[dag_node]['job_submission_file']
            if condor_job_submission_file not in condor_job_submission_files:
                condor_job_description = self.__parse_condor_job_submission_file(condor_job_submission_file)
                slurm_job_description = self.__condor_to_slurm_job_description(condor_job_description, dag_node)
                self.__write_slurm_job_submission_file(slurm_job_description, self.slurm_dag[dag_node]['job_submission_file'])
                condor_job_submission_files.append(condor_job_submission_file)


    def __parse_condor_job_submission_file(self, condor_job_submission_file):
        condor_job_description = {
            'executable': '',
            'arguments': '',
            'memoryusage': '',
            'memory': '',
            'cpus': '',
            'environment': [],
            'output': '',
            'error': '',
        }
        with open(condor_job_submission_file, 'r') as fd:
            for line in fd:
                if line.startswith('executable = '):
                    condor_job_description['executable'] = line.split(' = ' )[1].strip()
                elif line.startswith('arguments = '):
                    condor_job_description['arguments'] = line.strip('\n').replace('" ', '').replace(' "', '')[len('arguments = '):]
                elif line.startswith('+MemoryUsage = '):
                    condor_job_description['memoryusage'] = line.split(' = ')[1].strip('\n')
                elif line.startswith('request_memory = '):
                    condor_job_description['memory'] = line.split(' = ')[1].replace('MemoryUsage', condor_job_description['memoryusage']).strip('\n').replace('* 2 / 3 ) * 3 / 2', ')')
                elif line.startswith('request_cpus = '):
                    condor_job_description['cpus'] = line.split(' = ')[1].strip()
                elif line.startswith('environment = '):
                    condor_job_description['environment'] = line.strip('\n').replace('"', '').replace(';', '')[len('environment = '):].split()
                elif line.startswith('output = '):
                    condor_job_description['output'] = line.split(' = ')[1].strip()
                elif line.startswith('error = '):
                    condor_job_description['error'] = line.split(' = ')[1].strip()
        if condor_job_description['executable'] == '':
            msg  = 'Error parsing condor job submission file %s.\n' % (condor_job_submission_file)
            msg += 'No executable line found.'
            raise SyntaxError(msg)
        if condor_job_description['cpus'] == '':
            condor_job_description['cpus'] = self.condor_default_cpus
        if condor_job_description['memory'] == '':
            condor_job_description['memory'] = self.condor_default_memory
        return condor_job_description


    def __condor_to_slurm_job_description(self, condor_job_description, dag_node):
        slurm_job_description = {
            'cpus': condor_job_description['cpus'],
            'memory_per_cpu': self.__condor_to_slurm_memory(condor_job_description['memory'], condor_job_description['cpus']),
            'time': self.slurm_time_limit,
            'environment': self.__condor_to_slurm_environment(condor_job_description['environment'], dag_node),
            'output': self.__condor_to_slurm_output_file(condor_job_description['output']),
            'error': self.__condor_to_slurm_error_file(condor_job_description['error']),
            'payload': self.__condor_to_slurm_payload(condor_job_description['executable'], condor_job_description['arguments']),
        }
        return slurm_job_description


    def __condor_to_slurm_memory(self, condor_memory, cpus):
        memory = condor_memory
        if memory.endswith('MB'):
            memory = memory.replace('MB', '')
        elif memory.endswith('GB'):
            memory = memory.replace('GB', '') + ' * 1024'
        memory = int(eval(memory))
        memory_per_cpu = int(math.ceil(int(memory)/int(cpus)))
        return str(memory_per_cpu)


    def __condor_to_slurm_environment(self, condor_environment, dag_node):
        environment = 'ALL'
        for env in condor_environment:
            var, val = env.split('=')
            environment += ',%s' % (env)
        return environment


    def __condor_to_slurm_output_file(self, condor_output_file):
        if condor_output_file:
            output_file = condor_output_file.replace('$(cluster)', '%j').replace('-$(process)', '')
        else:
            output_file = 'slurm-%j.out'
        return output_file


    def __condor_to_slurm_error_file(self, condor_error_file):
        if condor_error_file:
            error_file = condor_error_file.replace('$(cluster)', '%j').replace('-$(process)', '')
        else:
            error_file = 'slurm-%j.err'
        return error_file


    def __condor_to_slurm_payload(self, condor_executable, condor_arguments):
        command = '%s %s' % (condor_executable, condor_arguments)
        if self.singularity_image:
            payload = \
"""\
export SINGULARITYENV_TMPDIR=${TMPDIR}
SINGULARITY_IMAGE=__SINGULARITY_IMAGE__
CMD="__COMMAND__"
# Passing the command to singularity via a pipe is to avoid this issue: https://github.com/hpcng/singularity/issues/5057
echo "${CMD}" | __SINGULARITY_COMMAND__\
"""
            singularity_command = 'singularity exec __SINGULARITY_OPTIONS__ ${SINGULARITY_IMAGE} bash -c "read -u 0 line; set -- \$line; exec \"\$@\""'
            if self.slurm_use_setsid:
                singularity_command = 'setsid ' + singularity_command
            bind_mount_dirs = ['${TMPDIR}']
            if self.singularity_bind_mount_dbus:
                bind_mount_dirs += ['/run/dbus', '/var/run/dbus']
            bind_mounts_opts_str = '-B ' + ' -B '.join(bind_mount_dirs)
            singularity_command = singularity_command.replace('__SINGULARITY_OPTIONS__', bind_mounts_opts_str)
            payload = payload \
                          .replace('__SINGULARITY_IMAGE__', self.singularity_image) \
                          .replace('__COMMAND__', command) \
                          .replace('__SINGULARITY_COMMAND__', singularity_command)
        else:
            payload = command
            if self.slurm_use_setsid:
                payload = 'setsid ' + payload
        return payload


    def __write_slurm_job_submission_file(self, slurm_job_description, slurm_job_submission_file):
        sbatch_directives_template = \
"""\
#SBATCH --cpus-per-task=__CPUS__
#SBATCH --mem-per-cpu=__MEMORY_PER_CPU__
#SBATCH --time=__TIME__
#SBATCH --export=__ENVIRONMENT__
#SBATCH --no-requeue
#SBATCH --output=__OUTPUT__
#SBATCH --error=__ERROR__\
"""
        sbatch_directives = sbatch_directives_template \
                      .replace('__CPUS__', slurm_job_description['cpus']) \
                      .replace('__MEMORY_PER_CPU__', slurm_job_description['memory_per_cpu']) \
                      .replace('__TIME__', slurm_job_description['time']) \
                      .replace('__ENVIRONMENT__', slurm_job_description['environment']) \
                      .replace('__OUTPUT__', slurm_job_description['output']) \
                      .replace('__ERROR__', slurm_job_description['error'])
        if self.slurm_partition:
            sbatch_directives += "\n#SBATCH --partition=%s" % (self.slurm_partition)
        if self.slurm_qos:
            sbatch_directives += "\n#SBATCH --qos=%s" % (self.slurm_qos)
        content_template = \
"""\
#!/bin/bash
#
__SBATCH_DIRECTIVES__

echo "======== Starting job (`date`) ========"
echo "Job id: ${SLURM_JOB_ID}"
echo "Job name: ${SLURM_JOB_NAME}"
echo "List of nodes allocated to the job: ${SLURM_JOB_NODELIST}"
echo "Batch node: ${SLURMD_NODENAME}"
echo "Working directory: `pwd`"

echo "==== Starting execution of payload (`date`) ===="
echo "------------------------- Begin payload output ------------------------"
(
__PAYLOAD__
)
pec=$?
echo "--------------------------- End payload output ------------------------"
echo "Payload exit code: ${pec}"
echo "==== Finished execution of payload (`date`) ===="

echo "======== Finished job (`date`) ========"
echo "Batch script exit code: ${pec}"
exit ${pec}
"""
        content = content_template \
                      .replace('__SBATCH_DIRECTIVES__', sbatch_directives) \
                      .replace('__PAYLOAD__', slurm_job_description['payload'])
        with open(slurm_job_submission_file, 'w') as fd:
            fd.write(content)



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import glob
import os
import re

from SlurmDagman import constants as C


def extract_rescue_number(dag_file):
    m = re.search('\.rescue([0-9]{3})$', dag_file)
    if m:
        return int(m.group(1))
    else:
        return 0


def is_rescue_dag_file(dag_file):
    return extract_rescue_number(dag_file) > 0


def get_dag_file_rootname(dag_file):
    if is_rescue_dag_file(dag_file):
        return dag_file[:-10]
    return dag_file


def get_highest_rescue_dag_file(dag_file):
    rescue_dag_files = get_all_rescue_dag_files(dag_file)
    if rescue_dag_files:
        return rescue_dag_files[-1]
    else:
        return dag_file


def get_all_rescue_dag_files(dag_file):
    dag_file_rootname = get_dag_file_rootname(dag_file)
    rescue_dag_files = glob.glob(dag_file_rootname + '.rescue[0-9][0-9][0-9]')
    if dag_file_rootname + '.rescue000' in rescue_dag_files:
        rescue_dag_files.remove(dag_file_rootname + '.rescue000')
    for rescue_dag_file in copy.copy(rescue_dag_files):
        rescue_number = extract_rescue_number(rescue_dag_file)
        if rescue_number > C.MAX_RESCUE_NUMBER:
            rescue_dag_files.remove(rescue_dag_file)
    return sorted(rescue_dag_files)


def build_rescue_dag_file_name(dag_file, rescue_number):
    dag_file_rootname = get_dag_file_rootname(dag_file)
    return dag_file_rootname + ('.rescue%03i' % (rescue_number))


def build_next_rescue_dag_file_name(dag_file):
    rescue_number = extract_rescue_number(dag_file)
    if rescue_number < C.MAX_RESCUE_NUMBER:
        return build_rescue_dag_file_name(dag_file, rescue_number + 1)
    else:
        return build_rescue_dag_file_name(dag_file, C.MAX_RESCUE_NUMBER)


def rename_rescue_dag_files(from_rescue_number, dag_file):
    rescue_dag_files = get_all_rescue_dag_files(dag_file)
    for rescue_dag_file in rescue_dag_files:
        rescue_number = extract_rescue_number(rescue_dag_file)
        if rescue_number >= from_rescue_number:
            os.rename(rescue_dag_file, rescue_dag_file + '.old')



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy


class Dag(object):

    def __init__(self, dag_file=''):
        super(Dag, self).__init__()
        self.dag_file = dag_file.strip()
        self.dag_nodes_appearance_order = {}
        self.dag = {}
        self.empty_node_template = {'job_submission_file': '', 'parents': [], 'done': False}
        self.max_retries = None
        self.no_retry_exit_codes = [0]


    def get_dag_file(self):
        return self.dag_file


    def set_dag_file(self, dag_file):
        self.dag_file = dag_file.strip()


    def set_dag(self, dag, copy_dag_nodes_appearance_order=False):
        if copy_dag_nodes_appearance_order and isinstance(dag, Dag):
            self.dag_nodes_appearance_order = copy.copy(dag.dag_nodes_appearance_order)
        for node in dag:
            self.dag[node] = copy.deepcopy(dag[node])


    def reset(self):
        for node in self.dag:
            self.reset_node(node)


    def reset_node(self, node):
        self.dag[node] = copy.deepcopy(self.empty_node_template)


    def get_nodes(self):
        return self.dag.keys()


    def parse(self):
        with open(self.dag_file, 'r') as fd:
            lines = fd.readlines()
            # Make a first pass to get all the dag nodes.
            for i, line in enumerate(lines):
                linestrip = line.strip()
                if linestrip.startswith('JOB '):
                    items = linestrip.split()
                    if len(items) < 3 or len(items) > 4 or (len(items) == 4 and items[3] != 'DONE'):
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Unexpected line format.\n"
                        msg += "Expected line format:\n"
                        msg += "JOB <node> <job-submission-file> [DONE]"
                        raise SyntaxError(msg)
                    node = items[1]
                    if node in self.dag_nodes_appearance_order:
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Node '%s' was already defined in line %i." % (node, self.dag_nodes_appearance_order[node])
                        raise SyntaxError(msg)
                    self.dag_nodes_appearance_order[node] = i
                    self.reset_node(node)
                    self.dag[node]['job_submission_file'] = items[2]
                    if len(items) == 4:
                        self.dag[node]['done'] = True
            # Make a second pass to find all the needed information about the dag nodes.
            for i, line in enumerate(lines):
                linestrip = line.strip()
                if linestrip.startswith('VARS '):
                    items = linestrip.split()
                    if len(items) < 3:
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Unexpected line format.\n"
                        msg += "Expected line format:\n"
                        msg += "VARS <node> <macro-1-key>=\"<macro-1-value>\" [<macro-2-key>=\"<macro-2-value>\" ...]\n"
                        msg += "(where macro values are allowed to contain white spaces)"
                        raise SyntaxError(msg)
                    node = items[1]
                    if node not in self.dag:
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Found a VARS line for node '%s', but there is no JOB line for this node." % (node)
                        raise SyntaxError(msg)
                    if 'vars' in self.dag[node]:
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Found a second VARS line for node '%s'.\n" % (node)
                        msg += "Only one VARS line can be specified per node."
                        raise SyntaxError(msg)
                    self.dag[node]['vars'] = ' '.join(items[2:])
                elif linestrip.startswith('PARENT '):
                    j = linestrip.find(' CHILD ')
                    if j == -1:
                        msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                        msg += "Unexpected line format.\n"
                        msg += 'Expected line format:\n'
                        msg += 'PARENT <parent-node-1> [<parent-node-2> ...] CHILD <child-node-1> [<child-node-2> ...]'
                        raise SyntaxError(msg)
                    parents = linestrip[:j][len('PARENT '):].replace(',', ' ').split()
                    children = linestrip[j+len(' CHILD '):].replace(',', ' ').split()
                    for parent in parents:
                        if parent not in self.dag:
                            msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                            msg += "There is no JOB line for parent node '%s'." % (parent)
                            raise SyntaxError(msg)
                    for child in children:
                        if child not in self.dag:
                            msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                            msg += "There is no JOB line for child node '%s'." % (child)
                            raise SyntaxError(msg)
                        for parent in parents:
                            if parent not in self.dag[child]['parents']:
                                self.dag[child]['parents'].append(parent)
                elif linestrip.startswith('RETRY '):
                    items = linestrip.split()
                    wrong_line_format_msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                    wrong_line_format_msg += "Unexpected line format.\n"
                    wrong_line_format_msg += "Expected line format:\n"
                    wrong_line_format_msg += "RETRY [<node> | ALL_NODES] <max-retries> [UNLESS-EXIT <exit-code1>[,<exit-code2>...]]\n"
                    wrong_line_format_msg += "(with max-retries a non-negative integer value and exit-code an integer value)"
                    if len(items) not in [3, 5]:
                        raise SyntaxError(wrong_line_format_msg)
                    try:
                        max_retries = int(items[2])
                        if max_retries < 0:
                            raise ValueError
                    except ValueError:
                        raise SyntaxError(wrong_line_format_msg)
                    if len(items) == 5:
                        if items[3] != 'UNLESS-EXIT':
                            raise SyntaxError(wrong_line_format_msg)
                        try:
                            no_retry_exit_codes = [int(i) for i in items[4].split(',')]
                        except ValueError:
                            raise SyntaxError(wrong_line_format_msg)
                    if items[1] == 'ALL_NODES':
                        if self.max_retries is not None:
                            msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                            msg += "Found a second RETRY ALL_NODES line.\n"
                            msg += "Only one RETRY ALL_NODES line can be specified."
                            raise SyntaxError(msg)
                        self.max_retries = max_retries
                        for node in self.dag:
                            self.dag[node].pop('max_retries', None)
                            if max_retries > 0:
                                self.dag[node]['retry_num'] = -1
                            else:
                                self.dag[node].pop('retry_num', None)
                        if len(items) == 5:
                            self.no_retry_exit_codes = no_retry_exit_codes
                            for node in self.dag:
                                self.dag[node].pop('no_retry_exit_codes', None)
                    else:
                        node = items[1]
                        if node not in self.dag:
                            msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                            msg += "Found a RETRY line for node '%s', but there is no JOB line for this node." % (node)
                            raise SyntaxError(msg)
                        if 'max_retries' in self.dag[node]:
                            msg  = "Error parsing dag file %s line %i.\n" % (self.dag_file, i)
                            msg += "Found a second RETRY line for node '%s'.\n" % (node)
                            msg += "Only one RETRY line can be specified per node."
                            raise SyntaxError(msg)
                        self.dag[node]['max_retries'] = max_retries
                        if max_retries > 0:
                            self.dag[node]['retry_num'] = -1
                        else:
                            self.dag[node].pop('retry_num', None)
                        if len(items) == 5:
                            self.dag[node]['no_retry_exit_codes'] = no_retry_exit_codes


    def get_max_retries(self, node):
        return self.dag[node].get('max_retries', self.max_retries)


    def get_no_retry_exit_codes(self, node):
        return self.dag[node].get('no_retry_exit_codes', self.no_retry_exit_codes)


    def write(self, dag_file='', use_dag_nodes_appearance_order=False, add_done_labels=False):
        dag_file = dag_file.strip()
        if dag_file == '':
            if self.dag_file != '':
                dag_file = self.dag_file
            else:
                raise ValueError('Dag file not specified.')
        if use_dag_nodes_appearance_order and self.dag_nodes_appearance_order:
            nodes = sorted(self.dag_nodes_appearance_order, key=self.dag_nodes_appearance_order.get)
        else:
            nodes = self.dag.keys()
        jobs_lines = []
        for node in nodes:
            job_line = 'JOB %s %s' % (node, self.dag[node]['job_submission_file'])
            if add_done_labels and self.dag[node]['done']:
                job_line += ' DONE'
            jobs_lines.append(job_line)
            if self.dag[node].get('vars', ''):
                vars_line = 'VARS %s %s' % (node, self.dag[node]['vars'])
                jobs_lines.append(vars_line)
            if 'max_retries' in self.dag[node]:
                retry_line = 'RETRY %s %i' % (node, self.dag[node]['max_retries'])
                jobs_lines.append(retry_line)
        dependency_lines = []
        for node in nodes:
            for parent in self.dag[node]['parents']:
                dependency_lines.append('PARENT %s CHILD %s' % (parent, node))
        retry_all_nodes_line = ['RETRY ALL_NODES %i' % (self.max_retries)] if self.max_retries is not None else [] 
        content = '\n'.join(jobs_lines + dependency_lines + retry_all_nodes_line)
        with open(dag_file, 'w') as fd:
            fd.write(content)


    def __str__(self):
        return str(self.dag)


    def __iter__(self):
        return iter(self.dag)


    def __next__(self, node):
        if node in self.dag:
            return self.dag[node]
        else:
            raise StopIteration


    def __getitem__(self, node):
        return self.dag[node]


    def __setitem__(self, node, value):
        self.dag[node] = value


    def __dict__(self):
        return self.dag



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import getpass
import os
import pwd
import subprocess
import sys

from argparse import ArgumentParser

from SlurmDagman import constants as C
from SlurmDagman.config.data.package import package_config
from SlurmDagman.dag.utils.rescue_dag import get_dag_file_rootname, get_highest_rescue_dag_file, extract_rescue_number, build_rescue_dag_file_name, rename_rescue_dag_files 


def exit(msg):
    print('ERROR: %s' % (msg))
    sys.exit(1)


options = {}

parser = ArgumentParser(description="Submit a DAG to SLURM via slurm_dagman")

parser.add_argument("--logfile",
                    dest = "logfile",
                    default = None,
                    help = "(slurm_dagman log file)")

parser.add_argument("--outfile",
                    dest = "outfile",
                    default = None,
                    help = "(slurm_dagman out file)")

parser.add_argument("--do-rescue-from",
                    type = int,
                    dest = "do_rescue_from",
                    default = 0,
                    help = "(run rescue DAG of given number)")

parser.add_argument("--no-rescue",
                    action = "store_true",
                    dest = "no_rescue",
                    default = False,
                    help = "(ignore rescue DAGs)")

user = getpass.getuser()
uid = pwd.getpwnam(user).pw_uid
parser.add_argument("--use-proxy",
                    action = "store_true",
                    dest = "use_proxy",
                    default = False,
                    help = "(add x509 user proxy to environment)")

parser.add_argument("--proxy-file",
                    dest = "proxy_file",
                    default = '/tmp/x509up_u%s' % (uid),
                    help = "(path to x509 user proxy)")

parser.add_argument("--sleep-time",
                    type = int,
                    dest = "sleep_time",
                    default = int(package_config.get_param('DAGMAN', 'sleep_time')),
                    help = "(time -in seconds- to sleep between two slurm dagman iterations)")

parser.add_argument("--max-jobs-queued",
                    type = int,
                    dest = "max_jobs_queued",
                    default = int(package_config.get_param('DAGMAN', 'max_jobs_queued')),
                    help = "(maximum number of jobs that can be put in the slurm queue)")

parser.add_argument("--max-jobs-submit",
                    type = int,
                    dest = "max_jobs_submit",
                    default = int(package_config.get_param('DAGMAN', 'max_jobs_submit')),
                    help = "(maximum number of jobs that can be submitted per iteration)")

parser.add_argument("--submit-wait-time",
                    type = int,
                    dest = "submit_wait_time",
                    default = int(package_config.get_param('DAGMAN', 'submit_wait_time')),
                    help = "(time -in seconds- to wait after a job submission)")

parser.add_argument("dagfile",
                    nargs = 1,
                    help = "(a DAG file)")

args = parser.parse_args()

if args.do_rescue_from < 0:
    parser.error('--do-rescue-from must be a non-negative interger (smaller than or equal to %s)' % (C.MAX_RESCUE_NUMBER))
if args.do_rescue_from > C.MAX_RESCUE_NUMBER:
    parser.error('--do-rescue-from must be smaller than or equal to %s' % (C.MAX_RESCUE_NUMBER))
if args.do_rescue_from > 0 and args.no_rescue:
    parser.error('--no-rescue can not be specified together with a non zero value of --do-rescue-from')
options['do_rescue_from'] = args.do_rescue_from

dag_file_rootname = get_dag_file_rootname(args.dagfile[0])

if args.do_rescue_from == 0:
    if args.no_rescue:
        options['dag_file'] = dag_file_rootname
    else:
        options['dag_file'] = get_highest_rescue_dag_file(dag_file_rootname)
else:
    options['dag_file'] = build_rescue_dag_file_name(dag_file_rootname, args.do_rescue_from)
options['dag_file'] = os.path.expanduser(os.path.abspath(options['dag_file']))

if not os.path.isfile(options['dag_file']):
    if args.do_rescue_from > 0:
        parser.error('--do-rescue-from %i specified, but rescue DAG file %s does not exist!' % (args.do_rescue_from, options['dag_file']))
    else:
        exit('DAG file %s does not exist!' % (options['dag_file']))

if args.do_rescue_from > 0:
    rename_rescue_dag_files(args.do_rescue_from + 1, options['dag_file'])

options['no_rescue'] = args.no_rescue
if args.logfile is None:
    options['logfile'] = dag_file_rootname + '.slurm_dagman.log'
else:
    options['logfile'] = args.logfile
options['logfile'] = os.path.expanduser(os.path.abspath(options['logfile']))
if args.outfile is None:
    options['outfile'] = dag_file_rootname + '.slurm_dagman.out'
else:
    options['outfile'] = args.outfile
options['outfile'] = os.path.expanduser(os.path.abspath(options['outfile']))

if args.use_proxy:
    options['proxy_file'] = os.path.expanduser(os.path.abspath(args.proxy_file))
    if not os.path.isfile(options['proxy_file']):
        exit('Proxy file %s does not exist!' % (options['proxy_file']))
else:
    options['proxy_file'] = None
options['use_proxy'] = args.use_proxy

options['sleep_time'] = max(args.sleep_time, 0)
options['max_jobs_queued'] = max(args.max_jobs_queued, 0)
options['max_jobs_submit'] = max(args.max_jobs_submit, 0)
options['submit_wait_time'] = max(args.submit_wait_time, 0)



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class ConfigurationError(configparser.Error):

    def __init__(self, msg, section, option=None):
        configparser.Error.__init__(self, msg)
        self.section = section
        self.option = option


class ConfigManager(object):

    def __init__(self):
        super(ConfigManager, self).__init__()    
        self.config = configparser.ConfigParser()


    def set_params(self, params):
        for section in params.keys():
            for option, value in params[section].items():
                self.set_param(section, option, value)


    def set_param(self, section, option, value):
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, option, value)


    def get_params(self, sections_and_options=None):
        params = OrderedDict()
        if sections_and_options is None:
            for section in self.config.sections():
                if section not in params:
                    params[section] = OrderedDict()
                for option in self.config.options(section):
                    params[section][option] = self.get_param(section, option)
        else:
            for section in sections_and_options.keys():
                if section not in params:
                    params[section] = OrderedDict()
                # options could be a list or a dict (in which case we only care about the keys)
                for option in sections_and_options[section]:
                    params[section][option] = self.get_param(section, option)
        return params


    def get_param(self, section, option, cast=None):
        if cast == 'int':
            return self.config.getint(section, option)
        if cast == 'float':
            return self.config.getfloat(section, option)
        if cast == 'boolean':
            return self.config.getboolean(section, option)
        return self.config.get(section, option)


    def load_from_file(self, possible_file_locations):
        self.config.read(possible_file_locations)


    def write_to_file(self, filename):
        with open(filename, 'w') as fd:
            self.config.write(fd)


    def validate(self, sections_and_options, must_exist=False):
        for section in self.config.sections():
            if section not in sections_and_options.keys():
                msg = 'Bad configuration: invalid section name %s' % (section)
                raise ConfigurationError(msg, section)
            for option in self.config.options(section):
                if option not in sections_and_options[section]:
                    msg = 'Bad configuration: invalid option name %s in section %s' % (option, section)
                    raise ConfigurationError(msg, section, option)
        if must_exist:
            for section in sections_and_options.keys():
                if not self.config.has_section(section):
                    msg = 'Bad configuration: missing section %s' % (section)
                    raise ConfigurationError(msg, section)
                for option in sections_and_options[section]:
                    if not self.config.has_option(section, option):
                        msg = 'Bad configuration: missing option %s in section %s' % (option, section)
                        raise ConfigurationError(msg, section, option)



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

DEFAULTS = OrderedDict()
DEFAULTS['DAGMAN'] = OrderedDict()
DEFAULTS['DAGMAN']['drain'] = 'no'
DEFAULTS['DAGMAN']['cancel'] = 'no'



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import getpass
import subprocess


def get_current_effective_user():
    return getpass.getuser()


def get_user_processes(user):
    cmd = ['ps', '-u', '%s' % (user), 'h', '-o', 'pid,command']
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    user_processes = {}
    for l in out.splitlines():
        pid, command = l.split(' ', 1)
        user_processes[pid] = command
    return user_processes


def find_slurm_dagman_pids(processes):
    pids = []
    for pid, command in processes.iteritems():
        if command.split(' ')[0] == 'python' and command.split(' ')[1].split('/')[-1] == 'slurm_dagman':
            pids.append(pid)
    return pids


def find_user_slurm_dagman_pids(user):
    user_processes = get_user_processes(user)
    user_slurm_dagman_pids = find_slurm_dagman_pids(user_processes)
    return user_slurm_dagman_pids


def find_slurm_dagman_pids_for_current_user():
    user = get_current_effective_user()
    user_slurm_dagman_pids = find_user_slurm_dagman_pids(user)
    return user, user_slurm_dagman_pids


def send_signal(pid, signal):
    cmd = ['kill', '-%s' % (signal), '%s' % (pid)]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    return out, err



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

DEFAULTS = OrderedDict()
DEFAULTS['SLURM'] = OrderedDict()
DEFAULTS['SLURM']['partition'] = ''
DEFAULTS['SLURM']['qos'] = ''
DEFAULTS['SLURM']['time_limit'] = '5-00:00:00'
DEFAULTS['SLURM']['scratch_dir'] = ''



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import site


def get_possible_package_config_locations():
    xdgConfigDirsDef = os.path.join(os.path.sep, 'etc', 'xdg')
    xdgConfigDirs = os.getenv('XDG_CONFIG_DIRS', xdgConfigDirsDef)
    if xdgConfigDirs.strip() == '':
        xdgConfigDirs = xdgConfigDirsDef
    xdgConfigHomeDef = os.path.join(os.path.expanduser('~'), '.config')
    xdgConfigHome = os.getenv('XDG_CONFIG_HOME', xdgConfigHomeDef)
    if xdgConfigHome.strip() == '':
        xdgConfigHome = xdgConfigHomeDef
    siteUserBase = site.USER_BASE
    virtualenv = os.getenv('VIRTUAL_ENV')
    configLocations = {
        'system'     : [os.path.join(os.path.sep, 'etc', 'SlurmDagman.conf'), os.path.join(os.path.sep, 'etc', 'SlurmDagman', 'SlurmDagman.conf')] + \
                       [os.path.join(xdgConfigDir.strip(), 'SlurmDagman.conf') for xdgConfigDir in xdgConfigDirs.split(':') if xdgConfigDir.strip()] + \
                       [os.path.join(xdgConfigDir.strip(), 'SlurmDagman', 'SlurmDagman.conf') for xdgConfigDir in xdgConfigDirs.split(':') if xdgConfigDir.strip()],
        'local'      : [os.path.join(siteUserBase, 'etc', 'SlurmDagman.conf'), os.path.join(siteUserBase, 'etc', 'SlurmDagman', 'SlurmDagman.conf')],
        'user'       : [os.path.join(xdgConfigHome, 'SlurmDagman.conf'), os.path.join(xdgConfigHome, 'SlurmDagman', 'SlurmDagman.conf')],
        'virtualenv' : [os.path.join(virtualenv, 'etc', 'SlurmDagman.conf'), os.path.join(virtualenv, 'etc', 'SlurmDagman', 'SlurmDagman.conf')] if virtualenv else [],
    }
    return configLocations['system'] + configLocations['local'] + configLocations['user'] + configLocations['virtualenv']




"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import datetime
import logging
import os
import re
import shutil
import subprocess
import time

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from SlurmDagman.config.data.process import process_config
from SlurmDagman.config.manager import ConfigManager, ConfigurationError
from SlurmDagman.config.utils.converters import text_to_bool
from SlurmDagman.dag import Dag
from SlurmDagman.dag.utils.rescue_dag import build_next_rescue_dag_file_name, get_dag_file_rootname


class Worker(object):

    def __init__(self, dag_file=None, outfile=None, proxy=None,
                 sleep_time=None, max_jobs_queued=None, max_jobs_submit=None, submit_wait_time=None):
        super(Worker, self).__init__()
        self.outfile = outfile
        self.__set_logging()
        self.__init_dag(dag_file)
        self.proxy = proxy
        self.__init_params()
        self.__init_process_config()
        self.process_config.set_params(process_config.get_params())
        self.process_config_file = get_dag_file_rootname(dag_file) + '.slurm_dagman.cfg'
        self.set_params(sleep_time, max_jobs_queued, max_jobs_submit, submit_wait_time)
        self.dag_done = {}
        self.num_nodes_total = 0
        self.num_nodes_done = 0
        self.num_nodes_queued = 0
        self.num_nodes_ready = 0
        self.num_nodes_unready = 0
        self.num_nodes_failed = 0
        self.start_time = datetime.datetime.now().isoformat().split('.')[0]
        self.wckey = 'slurm_dagman_%s' % (self.start_time.lower())
        self.queued_job_ids = []


    def __init_dag(self, dag_file=None):
        self.dag = Dag()
        if dag_file:
            self.set_dag_file(dag_file)


    def reset_dag(self):
        self.__init_dag()


    def set_dag_file(self, dag_file):
        self.dag.set_dag_file(dag_file)


    def __set_logging(self):
        if self.outfile is not None:
            logging.basicConfig(format="%(asctime)-15s %(message)s", datefmt='%m/%d/%y %H:%M:%S', filename=self.outfile, level=logging.DEBUG)


    def __init_params(self):
        self.sleep_time, self.max_jobs_queued, self.max_jobs_submit, self.submit_wait_time, \
        self.drain, self.cancel \
            = self.__get_config_params(process_config).values()


    def __init_process_config(self):
        self.process_config = ConfigManager()


    def __reset_process_config(self):
        self.__init_process_config()


    def __get_process_config_params(self, fallback=False, sanitize=True):
        return self.__get_config_params(self.process_config, fallback, sanitize)

 
    def __get_config_params(self, config, fallback=False, sanitize=True):
        params = OrderedDict()
        params['sleep_time'] = self.__get_config_param(config, 'DAGMAN', 'sleep_time', fallback, 'int')
        params['max_jobs_queued'] = self.__get_config_param(config, 'DAGMAN', 'max_jobs_queued', fallback, 'int')
        params['max_jobs_submit'] = self.__get_config_param(config, 'DAGMAN', 'max_jobs_submit', fallback, 'int')
        params['submit_wait_time'] = self.__get_config_param(config, 'DAGMAN', 'submit_wait_time', fallback, 'int')
        params['drain'] = self.__get_config_param(config, 'DAGMAN', 'drain', fallback, 'boolean')
        params['cancel'] = self.__get_config_param(config, 'DAGMAN', 'cancel', fallback, 'boolean')
        if sanitize:
            for name, value in params.items():
                params[name] = self.__replace_negative_int_by_zero(value)
        return params


    def __get_config_param(self, config, section, option, fallback=False, cast=None):
        try:
            param = config.get_param(section, option, cast)
        except Exception:
            if fallback:
                param = process_config.get_param(section, option, cast)
            else:
                param = None
        return param


    def __replace_negative_int_by_zero(self, param):
        if isinstance(param, int):
            return max(param, 0)
        return param
            

    def reset_params(self):
        self.__init_params()


    def set_params(self, sleep_time=None, max_jobs_queued=None, max_jobs_submit=None, submit_wait_time=None,
                         drain=None, cancel=None):
        if sleep_time is not None:
            self.sleep_time = self.__replace_negative_int_by_zero(sleep_time)
        if max_jobs_queued is not None:
            self.max_jobs_queued = self.__replace_negative_int_by_zero(max_jobs_queued)
        if max_jobs_submit is not None:
            self.max_jobs_submit = self.__replace_negative_int_by_zero(max_jobs_submit)
        if submit_wait_time is not None:
            self.submit_wait_time = self.__replace_negative_int_by_zero(submit_wait_time)
        if drain is not None:
            self.drain = drain
        if cancel is not None:
            self.cancel = cancel
        self.__set_process_config_params()
        self.__write_process_config()


    def __set_process_config_params(self):
        self.process_config.set_param('DAGMAN', 'sleep_time', self.sleep_time)
        self.process_config.set_param('DAGMAN', 'max_jobs_queued', self.max_jobs_queued)
        self.process_config.set_param('DAGMAN', 'max_jobs_submit', self.max_jobs_submit )
        self.process_config.set_param('DAGMAN', 'submit_wait_time', self.submit_wait_time )
        self.process_config.set_param('DAGMAN', 'drain', self.drain)
        self.process_config.set_param('DAGMAN', 'cancel', self.cancel)
 

    def __write_process_config(self):
        self.process_config.write_to_file(self.process_config_file)


    def __load_process_config(self):
        # Loading a config from file into memory is an update-like operation.
        # It can only overwrite existing config parameters in memory or define
        # new ones; it can not delete existing parameters in memory.
        if os.path.isfile(self.process_config_file):
            self.process_config.load_from_file([self.process_config_file])
            return True
        return False


    def __validate_process_config(self):
        try:
            self.process_config.validate(process_config.get_params(), True)
        except ConfigurationError:
            return False
        else:
            return True


    def __parse_process_config(self, log_changes=True):
        # Put the values of the local current process config in memory into
        # class variables. We don't want to fallback to use the global initial
        # process config if a parameter fails to be read from the local current
        # process config; instead we want to keep the current value in the class
        # variable. (A failure in reading a parameter from the local current
        # process config can happen when the parameter exists, but has an unexpected
        # type.) In case of such a failure, we get None for that parameter;
        # so we have to handle the case of a parameter being None. Finally,
        # we will return True if there is no None parameter and False otherwise.
        params = self.__get_process_config_params().values()
        sleep_time, max_jobs_queued, max_jobs_submit, submit_wait_time, \
        drain, cancel \
            = params[:]
        if log_changes:
            if sleep_time is not None and sleep_time != self.sleep_time:
                logging.info("Dag config change detected: sleep_time set to %s seconds" % (sleep_time))
            if max_jobs_queued is not None and max_jobs_queued != self.max_jobs_queued:
                logging.info("Dag config change detected: max_jobs_queued set to %s" % (max_jobs_queued))
            if max_jobs_submit is not None and max_jobs_submit != self.max_jobs_submit:
                logging.info("Dag config change detected: max_jobs_submit set to %s" % (max_jobs_submit))
            if submit_wait_time is not None and submit_wait_time != self.submit_wait_time:
                logging.info("Dag config change detected: submit_wait_time set to %s" % (submit_wait_time))
            if drain is not None and drain != self.drain:
                logging.info("Dag config change detected: drain set to %s" % (drain))
            if cancel is not None and cancel != self.cancel:
                logging.info("Dag config change detected: cancel set to %s" % (cancel))
        if sleep_time is not None:
            self.sleep_time = sleep_time
        if max_jobs_queued is not None:
            self.max_jobs_queued = max_jobs_queued
        if max_jobs_submit is not None:
            self.max_jobs_submit = max_jobs_submit
        if submit_wait_time is not None:
            self.submit_wait_time = submit_wait_time
        if drain is not None:
            self.drain = drain
        if cancel is not None:
            self.cancel = cancel
        return None not in params


    def __handle_process_config_loading_parsing_and_writing(self):
        rewrite = False
        self.__reset_process_config()
        if not self.__load_process_config():
            rewrite = True 
        elif not self.__parse_process_config() or not self.__validate_process_config():
            self.__reset_process_config()
            rewrite = True
        if rewrite:
            self.__set_process_config_params()
            self.__write_process_config()


    def __parse_dag_file(self):
        self.dag.parse()


    def __pre_execute_dag(self):
        nodes_all = self.dag.get_nodes()
        self.num_nodes_total = len(nodes_all)
        self.num_nodes_unready = len(nodes_all)
        for node in nodes_all:
            if self.dag[node]['done']:
                self.dag_done[node] = copy.deepcopy(self.dag[node])
                self.dag.__dict__().pop(node)
        del nodes_all
        nodes_done = self.dag_done.keys()
        self.num_nodes_done = len(nodes_done)
        self.num_nodes_unready -= self.num_nodes_done
        nodes_not_done = self.dag.get_nodes()
        for node in nodes_not_done:
            self.dag[node]['status'] = 0
            self.dag_done[node] = {'parents': []}
            parents = copy.copy(self.dag[node]['parents'])
            for parent in parents:
                if parent in nodes_done:
                    if parent not in self.dag_done[node]['parents']:
                        self.dag_done[node]['parents'].append(parent)
                    self.dag[node]['parents'].remove(parent)
            if not self.dag[node]['parents']:
                self.__mark_node_as_ready(node)


    def __pre_write_dag(self):
        for node in copy.copy(self.dag_done.keys()):
            if node not in self.dag:
                self.dag[node] = copy.deepcopy(self.dag_done[node])
            else:
                for parent in self.dag_done[node]['parents']:
                    if parent not in self.dag[node]['parents']:
                        self.dag[node]['parents'].append(parent)
            self.dag_done.pop(node)


    def write_rescue_dag_file(self):
        if self.num_nodes_done > 0:
            rescue_dag_file = build_next_rescue_dag_file_name(self.dag.get_dag_file())
            logging.info('Writing rescue dag file %s ...' % (rescue_dag_file))
            self.__write_dag_file(rescue_dag_file, True)
            logging.info('Rescue dag file saved.')


    def __write_dag_file(self, dag_file, add_done_labels=True):
        self.__pre_write_dag()
        self.dag.write(dag_file, use_dag_nodes_appearance_order=True, add_done_labels=add_done_labels)


    def __submit_ready_nodes(self):
        num_submitted_nodes = 0
        for node in self.dag:
            if self.max_jobs_queued > 0 and len(self.queued_job_ids) >= self.max_jobs_queued:
                break
            if self.dag[node]['status'] == 1: # node ready to be submitted
                # Submit a job
                job_id, error = self.__submit(self.dag[node]['job_submission_file'], node)
                if job_id is not None:
                    if 'retry_num' in self.dag[node] and self.dag[node]['retry_num'] > 0:
                        logging.info('Submitted node %s (retry number %i out of %i): %s' % (node, self.dag[node]['retry_num'], self.dag.get_max_retries(node), job_id))
                    else:
                        logging.info('Submitted node %s: %s' % (node, job_id))
                    self.__mark_node_as_queued(node, job_id)
                    self.queued_job_ids.append(job_id)
                    num_submitted_nodes += 1
                else:
                    self.__mark_node_as_queued(node)
                    if 'retry_num' in self.dag[node] and self.dag[node]['retry_num'] > 0:
                        logging.error('Failed to submit node %s (retry number %i out of %i)' % (node, self.dag[node]['retry_num'], self.dag.get_max_retries(node)))
                    else:
                        logging.error('Failed to submit node %s' % (node))
                    if 'retry_num' in self.dag[node] and self.dag[node]['retry_num'] < self.dag.get_max_retries(node):
                        self.__mark_node_as_ready(node)
                        logging.info('Node %s will be retried.' % (node))
                    else:
                        self.__mark_node_as_failed(node)
                    if error is not None:
                        logging.debug('Error was: %s' % (error))
                if self.max_jobs_submit > 0 and num_submitted_nodes >= self.max_jobs_submit:
                    break
                if self.submit_wait_time > 0:
                    time.sleep(self.submit_wait_time)


    def __submit(self, job_submission_file, node):
        with open(job_submission_file + '.tmp', 'w') as fd1:
            with open(job_submission_file, 'r') as fd2:
                for line in fd2:
                    fd1.write(self.__replace_vars(line, node))
        cmd = ['sbatch', '--job-name=%s' % (node), '--wckey=%s' % (self.wckey), job_submission_file + '.tmp']
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode('utf-8').strip()
        err = err.decode('utf-8').strip()
        os.remove(job_submission_file + '.tmp')
        if err or not out:
            return None, err
        job_id = out.strip('\n').split()[-1]
        return job_id, None


    def __handle_proxy(self):
        if self.proxy is not None:
            destdir = os.path.expanduser('~')
            if not os.path.exists(destdir):
                os.mkdir(destdir)
            elif not os.path.isdir(destdir):
                raise OSError('Error: %s exists and it is not a directory.' % (destdir)) 
            shutil.copy(self.proxy, os.path.join(destdir, '.x509-user-proxy'))
            self.proxy = os.path.join(destdir, '.x509-user-proxy')
            os.environ['X509_USER_PROXY'] = self.proxy


    def __replace_vars(self, string, node):
        startindex = 0
        while True:
            i = string.find('$(', startindex)
            j = string.find(')', max(startindex, i))
            startindex = j+1
            if not (i > -1 and j > i):
                break
            macrokey = string[i+2:j]
            m = self.dag[node].get('vars', '').split(macrokey+'="')
            if len(m) == 2:
                n = m[1].find('"')
                if n > -1:
                    macrovalue = m[1][:n]
                    string = string[:i] + macrovalue + string[j+1:]
                    startindex += (len(macrovalue) - (len(macrokey) + 3))
                else:
                    string = string[:i] + string[j+1:]
                    startindex -= (len(macrokey) + 3)
            else:
                string = string[:i] + string[j+1:]
                startindex -= (len(macrokey) + 3)
        return string


    def __sacct(self):
        cmd = ['sacct', '--noheader', '-P', '--wckeys=%s' % (self.wckey), '--format=JobID,JobName,State,NodeList,ExitCode', '--jobs=%s' % (','.join(self.queued_job_ids)), '--starttime=%s' % (self.start_time)]
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode('utf-8').strip().split('\n')
        err = err.decode('utf-8').strip().split('\n')
        out = [l for l in out if l and l.find('|batch|') == -1 and l.find('|extern|') == -1]
        return out, err


    def __squeue(self):
        cmd = ['squeue', '--noheader', '--format="%i|%T|%w|%N"']
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode('utf-8').strip().split('\n')
        err = err.decode('utf-8').strip().split('\n')
        out = [l for l in out if l and l.find('|%s|' % (self.wckey)) > 0]
        return out, err


    def __monitor(self):
        num_nodes_running = 0
        num_nodes_pending = 0
        num_nodes_unknown = 0
        sacct_out, _ = self.__sacct()
        squeue_out, _ = self.__squeue()
        squeue_result = {}
        for job in squeue_out:
            job_id, status, wckey, computing_node = job.split('|')
            squeue_result[job_id] = {'status': status, 'computing_node': computing_node}
        nodes_done = set()
        for job in sacct_out:
            job_id, job_name, status, computing_node, exit_code = job.split('|')
            try:
                exit_code = int(exit_code.split(':')[0])
            except ValueError:
                exit_code = 'undefined'
            if job_id in squeue_result:
                status = squeue_result[job_id]['status']
                computing_node = squeue_result[job_id]['computing_node']
            #if self.dag[job_name]['status'] == 2:
            if status == 'PENDING':
                num_nodes_pending += 1
            elif status in ['RUNNING', 'COMPLETING']:
                num_nodes_running += 1
            elif status == 'COMPLETED':
                logging.info('Node %s completed' % (job_name))
                self.__mark_node_as_done(job_name)
                nodes_done.add(job_name)
                self.queued_job_ids.remove(job_id)
            elif status in ['RESIZING', 'REQUEUED', 'REVOKED', 'SUSPENDED']:
                num_nodes_unknown += 1
            else:
                if status == 'FAILED':
                    logging.info('Node %s failed (exit code %s)' % (job_name, exit_code))
                else:
                    logging.info('Node %s in status %s (exit_code %s) assumed to have failed' % (job_name, status, exit_code))
                if 'retry_num' in self.dag[job_name] and self.dag[job_name]['retry_num'] < self.dag.get_max_retries(job_name) and exit_code not in self.dag.get_no_retry_exit_codes(job_name):
                    self.__mark_node_as_ready(job_name)
                    logging.info('Node %s will be retried' % (job_name))
                else:
                    self.__mark_node_as_failed(job_name)
                self.queued_job_ids.remove(job_id)
        if nodes_done:
            self.__fix_parents(nodes_done)
        return num_nodes_running, num_nodes_pending, num_nodes_unknown


    def __fix_parents(self, nodes_done):
        for node in self.dag:
            if self.dag[node]['status'] == 0:
                for node_done in nodes_done:
                    if node_done in self.dag[node]['parents']:
                        if node not in self.dag_done: 
                            self.dag_done[node] = {'parents': []}
                        self.dag_done[node]['parents'].append(node_done)
                        self.dag[node]['parents'].remove(node_done)
                        if not self.dag[node]['parents']:
                            self.__mark_node_as_ready(node)
                            break


    def __mark_node_as_ready(self, node):
        self.dag[node]['status'] = 1
        self.num_nodes_ready += 1
        if 'retry_num' in self.dag[node]:
            self.dag[node]['retry_num'] += 1
            if self.dag[node]['retry_num'] > 0:
                self.num_nodes_queued -= 1
            else:
                self.num_nodes_unready -= 1
        else:
            self.num_nodes_unready -= 1


    def __mark_node_as_queued(self, node, job_id=None):
        if job_id is not None:
            self.dag[node]['job_id'] = job_id
        self.dag[node]['status'] = 2
        self.num_nodes_queued += 1
        self.num_nodes_ready -= 1


    def __mark_node_as_done(self, node):
        self.dag[node]['status'] = 3
        self.dag[node]['done'] = True
        self.num_nodes_done += 1
        self.num_nodes_queued -= 1
        for parent in self.dag_done[node]['parents']:
            if parent not in self.dag[node]['parents']:
                self.dag[node]['parents'].append(parent)
        self.dag_done.pop(node)
        self.dag_done[node] = copy.deepcopy(self.dag[node])
        self.dag.__dict__().pop(node)


    def __mark_node_as_failed(self, node):
        self.dag[node]['status'] = 4
        self.num_nodes_failed += 1
        self.num_nodes_queued -= 1


    def __execute_dag(self):
        self.__pre_execute_dag()
        logging.info('******************************************************')
        logging.info('** PID = %s' % (os.getpid()))
        if os.getppid():
            logging.info('** Parent PID = %s' % (os.getppid()))
        logging.info('******************************************************')
        logging.info('DAG file: %s' % (self.dag.get_dag_file()))
        logging.info('DAGMan config file for this DAG: %s' % (self.process_config_file))
        logging.info('Sleep time between iterations: %s secs' % (self.sleep_time))
        logging.info('Will submit jobs with wckey=%s' % (self.wckey))
        while True:
            num_nodes_running, num_nodes_pending, num_nodes_unknown = self.__monitor()
            logging.info('Of %i nodes total:' % (self.num_nodes_total))
            logging.info('  Done   Queued    Ready   Un-Ready   Failed')
            logging.info('   ===      ===      ===        ===      ===')
            logging.info('%6s   %6s   %6s     %6s   %6s' % (self.num_nodes_done, self.num_nodes_queued, self.num_nodes_ready, self.num_nodes_unready, self.num_nodes_failed))
            if self.num_nodes_queued > 0:
                logging.info('Of %i nodes queued: %i running, %i pending, %i other' % (self.num_nodes_queued, num_nodes_running, num_nodes_pending, num_nodes_unknown))
            if self.num_nodes_ready > 0 and not self.drain and not self.cancel:
                self.__submit_ready_nodes()
            else:
                if self.num_nodes_done == self.num_nodes_total:
                    logging.info('DAG completed successfully.')
                    return 0
                elif self.num_nodes_queued == 0:
                    if self.num_nodes_failed:
                        logging.info('DAG failed.')
                        return 1
                    else:
                        logging.info('DAG stopped.')
                        return 2
            if self.cancel:
                return 2
            if self.sleep_time > 0:
                time.sleep(self.sleep_time)
            self.__handle_process_config_loading_parsing_and_writing()


    def __cancel_dag(self):
        max_num_cancel_retries = 5
        cancel_retry_num = 1
        while True:
            cmd = ['scancel', '--wckey=%s' % (self.wckey)]
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
            time.sleep(60)
            squeue_out, squeue_err = self.__squeue()
            if squeue_out or squeue_err:
                if cancel_retry_num > max_num_cancel_retries:
                    msg  = 'After %i trials to cancel the queued DAG nodes, %i nodes are still queued.' % (cancel_retry_num, len(squeue_out))
                    msg += ' Aborting the cancel.'
                    msg += ' This is the output of squeue:'
                    msg += '\nStdout:\n%s' % ('\n'.join(squeue_out))
                    if squeue_err:
                        msg += '\nStderr:\n%s' % ('\n'.join(squeue_err))
                    raise Exception(msg)
                cancel_retry_num += 1
            else:
                break


    def run(self):
        self.__handle_proxy()
        try:
            self.__parse_dag_file()
        except Exception as e:
            logging.exception('Failure parsing DAG file.')
            raise
        try:
            rc = self.__execute_dag()
        except Exception as e:
            logging.exception('Failure executing DAG.')
            raise
        if self.cancel and rc != 0:
            self.try_to_terminate_and_write_rescue_dag_file()
            rc = -1
        return rc


    def terminate(self):
        if self.num_nodes_queued > 0:
            try:
                logging.info('Cancelling queued DAG nodes.')
                self.__cancel_dag()
                logging.info('Queued DAG nodes cancelled.')
            except Exception as e:
                logging.exception('Failure cancelling queued DAG nodes.')
                raise


    def try_to_terminate_and_write_rescue_dag_file(self):
        try:
            self.terminate()
        except Exception:
            logging.warning('Failed to terminate DAG. Will write rescue DAG file with currently finished nodes.')
            pass
        try:
            self.write_rescue_dag_file()
        except Exception:
            logging.exception('Failed to write rescue DAG.')
            raise



""" Setuptools-based setup module for SlurmDagman """

from setuptools import setup, find_packages
import os, os.path

from io import open
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

from SlurmDagman import __version__ as version

pkgName = "SlurmDagman"

setup(
    name=pkgName,
    version=version,

    description="Application to run a DAG with SLURM.",
    long_description=long_description,
    long_description_content_type='text/markdown',

    url="https://github.com/AndresTanasijczuk/SlurmDagman",

    author="Andres Tanasijczuk (Université catholique de Louvain)",
    author_email="cp3-support@listes.uclouvain.be",
    license="GPL-3.0-or-later",

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Distributed Computing',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords=['slurm','dag'],

    package_dir={'' : 'lib'},
    packages=find_packages('lib'),
    scripts=[os.path.join(root, item) for root, subFolder, files in os.walk('bin') for item in files],
    # This works together with MANIFEST.in for sdist.
    #include_package_data=True,
    # This is for bdist_wheel.
    #data_files=[('./libexec',['scripts/slurm_dagman'])],
    data_files=[(os.path.join('etc', pkgName), [os.path.join('etc', 'SlurmDagman.conf')])],
)



"""
Copyright (C) 2020  Universite catholique de Louvain, Belgium.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy

from SlurmDagman.config.defaults.common import DEFAULTS as common_config_defaults
from SlurmDagman.config.defaults.package import DEFAULTS as package_config_defaults
from SlurmDagman.config.manager import ConfigManager
from SlurmDagman.config.utils.package import get_possible_package_config_locations


package_config = ConfigManager()
package_config.set_params(package_config_defaults)
package_config.set_params(common_config_defaults)
package_config.load_from_file(get_possible_package_config_locations())
defaults = {} 
for d in [package_config_defaults, common_config_defaults]:
    for section in d.keys():
        if section not in defaults:
            defaults[section] = {}
        for option in d[section].keys():
            defaults[section][option] = None
package_config.validate(defaults)
